=head1 NAME

DynGig::Multiplex - A collection of multiplex modules

=cut
package DynGig::Multiplex;

=head1 VERSION

Version 1.02

=cut
our $VERSION = '1.02';

=head1 MODULES

=head2 DynGig::Multiplex::CMD 

Multiplexed Fork Client

=head2 DynGig::Multiplex::TCP 

Multiplexed TCP Client

=head1 AUTHOR

Kan Liu

=head1 COPYRIGHT and LICENSE

Copyright (c) 2010. Kan Liu

This program is free software; you may redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1;

__END__
